@@ -9,6 +9,7 @@
 -- Stability   : stable
 module GitLab
   ( runGitLab,
+    runGitLabPassPrompt,
     runGitLabDbg,
     runGitLabWithManager,
     module GitLab.Types,
@@ -38,6 +39,8 @@ where
 
 import Control.Monad.IO.Class
 import Control.Monad.Trans.Reader
+import Data.Default.Class
+import qualified Data.Text as T
 import GitLab.API.Boards
 import GitLab.API.Branches
 import GitLab.API.Commits
@@ -60,7 +63,6 @@ import GitLab.SystemHooks.GitLabSystemHooks
 import GitLab.SystemHooks.Rules
 import GitLab.SystemHooks.Types
 import GitLab.Types
-import Network.Connection (TLSSettings (..))
 import Network.HTTP.Conduit
 import Network.HTTP.Types.Status
 import System.IO
@@ -85,16 +87,34 @@ import System.IO
 runGitLab :: GitLabServerConfig -> GitLab a -> IO a
 runGitLab cfg action = do
   liftIO $ hSetBuffering stdout LineBuffering
-  let settings = mkManagerSettings (TLSSettingsSimple True False False) Nothing
+  let settings = mkManagerSettings def Nothing
   manager <- liftIO $ newManager settings
   runGitLabWithManager manager cfg action
 
+-- | The same as 'runGitLab', except that it prompts for a GitLab
+-- access token before running the GitLab action.
+--
+-- In this case you can just use 'defaultGitLabServer' with no
+-- modification of the record field values, because these values will
+-- be asked for at runtime:
+--
+-- > runGitLabPassPrompt defaultGitLabServer myGitLabProgram
+runGitLabPassPrompt :: GitLabServerConfig -> GitLab a -> IO a
+runGitLabPassPrompt cfg action = do
+  liftIO $ hSetBuffering stdout NoBuffering
+  liftIO (putStr "Enter GitLab server URL\n> ")
+  hostUrl <- getLine
+  liftIO (putStr "Enter GitLab access token\n> ")
+  pass <- getLine
+  runGitLab (cfg {url = T.pack hostUrl, token = AuthMethodToken (T.pack pass)}) action
+
 -- | The same as 'runGitLab', except that it also takes a connection
 -- manager as an argument.
 runGitLabWithManager :: Manager -> GitLabServerConfig -> GitLab a -> IO a
-runGitLabWithManager manager cfg action = do
+runGitLabWithManager manager cfg (GitLabT action) = do
   -- test the token access
-  tokenTest <- runReaderT gitlabVersion (GitLabState cfg manager)
+  let (GitLabT versionCheck) = gitlabVersion
+  tokenTest <- runReaderT versionCheck (GitLabState cfg manager)
   case tokenTest of
     Left response ->
       case responseStatus response of
@@ -108,6 +128,8 @@ runGitLabWithManager manager cfg action = do
 -- liftIO. Cannot speak to a GitLab server. Only useful for the
 -- gitlab-haskell tests.
 runGitLabDbg :: GitLab a -> IO a
-runGitLabDbg action = do
+runGitLabDbg (GitLabT action) = do
   liftIO $ hSetBuffering stdout LineBuffering
-  runReaderT action undefined
+  manager <- liftIO $ newManager (mkManagerSettings def Nothing)
+  let cfg = GitLabServerConfig {url = "", token = AuthMethodToken "", retries = 1, debugSystemHooks = Nothing}
+  runReaderT action (GitLabState cfg manager)
