@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE DuplicateRecordFields #-}
 {-# LANGUAGE GADTs #-}
@@ -46,6 +47,7 @@ module GitLab.SystemHooks.Types
     CommitAuthorEvent (..),
     Visibility (..),
     MergeRequestEvent (..),
+    Label (..),
     MergeRequestChanges (..),
     MergeRequestChange (..),
     ObjectAttributes (..),
@@ -56,8 +58,8 @@ module GitLab.SystemHooks.Types
 where
 
 import Data.Aeson
-import qualified Data.ByteString.Lazy.Char8 as BSL
 import Data.Text (Text)
+import qualified Data.Text.Encoding as T
 import Data.Typeable
 import GHC.Generics
 import GitLab.Types
@@ -73,9 +75,9 @@ class (FromJSON a) => SystemHook a where
   matchIf :: String -> (a -> GitLab Bool) -> (a -> GitLab ()) -> Rule
 
 -- | Parse JSON data into GitLab events.
-parseEvent :: (FromJSON a) => String -> Maybe a
-parseEvent string =
-  case eitherDecode (BSL.pack string) of
+parseEvent :: (FromJSON a) => Text -> Maybe a
+parseEvent eventText =
+  case eitherDecodeStrict (T.encodeUtf8 eventText) of
     Left _error -> Nothing
     Right event -> Just event
 
@@ -466,7 +468,8 @@ data Push = Push
     push_checkout_sha :: Text,
     push_user_id :: Int,
     push_user_name :: Text,
-    push_user_email :: Text,
+    push_user_username :: Maybe Text,
+    push_user_email :: Maybe Text,
     push_user_avatar :: Text,
     push_project_id :: Int,
     push_project :: ProjectEvent,
@@ -519,7 +522,7 @@ data RepositoryUpdate = RepositoryUpdate
 -- | A project event.
 data ProjectEvent = ProjectEvent
   { projectEvent_name :: Text,
-    projectEvent_description :: Text,
+    projectEvent_description :: Maybe Text,
     projectEvent_web_url :: Text,
     projectEvent_avatar_url :: Maybe Text,
     projectEvent_git_ssh_url :: Text,
@@ -548,7 +551,7 @@ data ProjectChanges = ProjectChanges
 data RepositoryEvent = RepositoryEvent
   { repositoryEvent_name :: Text,
     repositoryEvent_url :: Text,
-    repositoryEvent_description :: Text,
+    repositoryEvent_description :: Maybe Text,
     repositoryEvent_homepage :: Maybe Text,
     -- these three not in the merge_request event example
     -- in the GitLab documentation. Is the merge_request documentation
@@ -588,39 +591,58 @@ data MergeRequestEvent = MergeRequestEvent
     mergeRequest_user :: UserEvent,
     mergeRequest_project :: ProjectEvent,
     mergeRequest_object_attributes :: ObjectAttributes,
-    mergeRequest_labels :: Maybe [Text],
+    mergeRequest_labels :: Maybe [Label],
     mergeRequest_changes :: MergeRequestChanges,
     mergeRequest_repository :: RepositoryEvent
   }
   deriving (Typeable, Show, Eq, Generic)
 
-data MergeRequestChanges = MergeRequestChanges
-  { mergeRequestChanges_author_id :: MergeRequestChange Int,
-    mergeRequestChanges_created_at :: MergeRequestChange Text,
-    mergeRequestChanges_description :: MergeRequestChange Text,
-    mergeRequestChanges_id :: MergeRequestChange Int,
-    mergeRequestChanges_iid :: MergeRequestChange Int,
-    mergeRequestChanges_source_branch :: MergeRequestChange Text,
-    mergeRequestChanges_source_project_id :: MergeRequestChange Int,
-    mergeRequestChanges_target_branch :: MergeRequestChange Text,
-    mergeRequestChanges_target_project_id :: MergeRequestChange Int,
-    mergeRequestChanges_title :: MergeRequestChange Text,
-    mergeRequestChanges_updated_at :: MergeRequestChange Text
+-- | Label associated with a merge request
+data Label = Label
+  { label_id :: Maybe Int,
+    label_title :: Maybe Text,
+    label_color :: Maybe Text,
+    label_project_id :: Maybe Int,
+    label_created_at :: Maybe Text, -- TODO date from e.g. "2013-12-03T17:15:43Z"
+    label_updated_at :: Maybe Text, -- TODO date
+    label_template :: Maybe Bool,
+    label_description :: Maybe Text,
+    label_type :: Maybe Text, -- TODO type from "ProjectLabel"
+    label_group_id :: Maybe Int
   }
   deriving (Typeable, Show, Eq, Generic)
 
+-- | Changes that a merge request will make
+data MergeRequestChanges = MergeRequestChanges
+  { mergeRequestChanges_author_id :: Maybe (MergeRequestChange Int),
+    mergeRequestChanges_created_at :: Maybe (MergeRequestChange Text),
+    mergeRequestChanges_description :: Maybe (MergeRequestChange Text),
+    mergeRequestChanges_id :: Maybe (MergeRequestChange Int),
+    mergeRequestChanges_iid :: Maybe (MergeRequestChange Int),
+    mergeRequestChanges_source_branch :: Maybe (MergeRequestChange Text),
+    mergeRequestChanges_source_project_id :: Maybe (MergeRequestChange Int),
+    mergeRequestChanges_target_branch :: Maybe (MergeRequestChange Text),
+    mergeRequestChanges_target_project_id :: Maybe (MergeRequestChange Int),
+    mergeRequestChanges_title :: Maybe (MergeRequestChange Text),
+    mergeRequestChanges_updated_at :: Maybe (MergeRequestChange Text)
+  }
+  deriving (Typeable, Show, Eq, Generic)
+
+-- | The change between for a given GitLab data field a merge request
+-- will make
 data MergeRequestChange a = MergeRequestChange
   { mergeRequestChange_previous :: Maybe a,
     mergeRequestChange_current :: Maybe a
   }
   deriving (Typeable, Show, Eq, Generic)
 
+-- | Attributes associated with a merge request
 data ObjectAttributes = ObjectAttributes
   { objectAttributes_id :: Int,
     objectAttributes_target_branch :: Text,
     objectAttributes_source_branch :: Text,
     objectAttributes_source_project_id :: Int,
-    objectAttributes_author_id :: Int,
+    objectAttributes_author_id :: Maybe Int,
     objectAttributes_assignee_id :: Maybe Int,
     objectAttributes_assignee_ids :: Maybe [Int],
     objectAttributes_title :: Text,
@@ -635,14 +657,14 @@ data ObjectAttributes = ObjectAttributes
     objectAttributes_description :: Text,
     objectAttributes_updated_by_id :: Maybe Int,
     objectAttributes_merge_error :: Maybe Text,
-    objectAttributes_merge_params :: MergeParams,
-    objectAttributes_merge_when_pipeline_succeeds :: Bool,
+    objectAttributes_merge_params :: Maybe MergeParams,
+    objectAttributes_merge_when_pipeline_succeeds :: Maybe Bool,
     objectAttributes_merge_user_id :: Maybe Int,
     objectAttributes_merge_commit_sha :: Maybe Text,
     objectAttributes_deleted_at :: Maybe Text,
     objectAttributes_in_progress_merge_commit_sha :: Maybe Text,
     objectAttributes_lock_version :: Maybe Int,
-    objectAttributes_time_estimate :: Int,
+    objectAttributes_time_estimate :: Maybe Int,
     objectAttributes_last_edited_at :: Maybe Text,
     objectAttributes_last_edited_by_id :: Maybe Int,
     objectAttributes_head_pipeline_id :: Maybe Int,
@@ -652,18 +674,20 @@ data ObjectAttributes = ObjectAttributes
     objectAttributes_target :: ProjectEvent,
     objectAttributes_last_commit :: CommitEvent,
     objectAttributes_work_in_progress :: Bool,
-    objectAttributes_total_time_spent :: Int,
+    objectAttributes_total_time_spent :: Maybe Int,
     objectAttributes_human_total_time_spent :: Maybe Int,
     objectAttributes_human_time_estimate :: Maybe Int,
     objectAttributes_action :: Maybe Text
   }
   deriving (Typeable, Show, Eq, Generic)
 
-data MergeParams = MergeParams
+-- | Merge parameters associated with a merge request
+newtype MergeParams = MergeParams
   { mergeParams_force_remove_source_branch :: Maybe Text
   }
   deriving (Typeable, Show, Eq, Generic)
 
+-- | User associated with a merge request
 data UserEvent = UserEvent
   { userEvent_name :: Text,
     userEvent_username :: Text,
@@ -698,13 +722,6 @@ data ProjectAction
   | MergeRequested
   deriving (Show, Eq)
 
--- |  Project visibility.
-data Visibility
-  = Public
-  | Private
-  | Internal
-  deriving (Show, Eq)
-
 instance FromJSON ProjectCreate where
   parseJSON =
     withObject "ProjectCreate" $ \v -> do
@@ -1154,6 +1171,7 @@ instance FromJSON Push where
                 <*> v .: "checkout_sha"
                 <*> v .: "user_id"
                 <*> v .: "user_name"
+                <*> v .: "user_username"
                 <*> v .: "user_email"
                 <*> v .: "user_avatar"
                 <*> v .: "project_id"
@@ -1408,6 +1426,32 @@ instance (FromJSON a) => FromJSON (MergeRequestChange a) where
           }
       )
 
+instance FromJSON Label where
+  parseJSON = withObject "Label" $ \obj -> do
+    labelId <- obj .:? "id"
+    labelTitle <- obj .:? "title"
+    labelColor <- obj .:? "color"
+    labelProjectId <- obj .:? "project_id"
+    labelCreatedAt <- obj .:? "created_at"
+    labelUpdatedAt <- obj .:? "updated_at"
+    labelTemplate <- obj .:? "template"
+    labelDescription <- obj .:? "description"
+    labelType <- obj .:? "type"
+    labelGroupId <- obj .:? "group_id"
+    return $
+      Label
+        { label_id = labelId,
+          label_title = labelTitle,
+          label_color = labelColor,
+          label_project_id = labelProjectId,
+          label_created_at = labelCreatedAt,
+          label_updated_at = labelUpdatedAt,
+          label_template = labelTemplate,
+          label_description = labelDescription,
+          label_type = labelType,
+          label_group_id = labelGroupId
+        }
+
 instance FromJSON ProjectAction where
   parseJSON (String "project_create") = return ProjectCreated
   parseJSON (String "project_destroy") = return ProjectDestroyed
@@ -1433,13 +1477,4 @@ instance FromJSON ProjectAction where
   parseJSON (String "tag_push") = return TagPushed
   parseJSON (String "repository_update") = return RepositoryUpdated
   parseJSON (String "merge_request") = return MergeRequested
-  parseJSON s = error ("unexpected system hook event: " <> show s)
-
-instance FromJSON Visibility where
-  parseJSON (String "public") = return Public
-  parseJSON (String "private") = return Private
-  parseJSON (String "internal") = return Internal
-  parseJSON (Number 0) = return Private
-  parseJSON (Number 10) = return Internal
-  parseJSON (Number 20) = return Public
-  parseJSON n = error (show n)
+  parseJSON s = fail ("unexpected system hook event: " <> show s)
