@@ -10,15 +10,17 @@ module GitLab.WebRequests.GitLabWebCalls
     gitlabPut,
     gitlabDelete,
     gitlabUnsafe,
+    gitlabGetByteStringResponse,
   )
 where
 
 import qualified Control.Exception as Exception
 import Control.Monad.IO.Class
-import Control.Monad.Trans.Reader
+import qualified Control.Monad.Reader as MR
 import Data.Aeson
-import Data.ByteString
+import qualified Data.ByteString as BS
 import qualified Data.ByteString.Lazy as BSL
+import Data.Maybe (isJust)
 import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Text.Encoding as T
@@ -33,7 +35,7 @@ newtype GitLabException = GitLabException String
 
 instance Exception.Exception GitLabException
 
-type GitLabParam = (ByteString, Maybe ByteString)
+type GitLabParam = (BS.ByteString, Maybe BS.ByteString)
 
 gitlabGetOne ::
   (FromJSON a) =>
@@ -105,14 +107,14 @@ gitlabPut urlPath params = do
         params
 
 gitlabDelete ::
+  FromJSON a =>
   -- | the URL to post to
   Text ->
-  GitLab (Either (Response BSL.ByteString) (Maybe ()))
-gitlabDelete urlPath = do
-  result <- request
-  case result of
-    Right (Just _) -> return (Right (Just ()))
-    x -> return x
+  -- | the data to post
+  [GitLabParam] ->
+  GitLab (Either (Response BSL.ByteString) (Maybe a))
+gitlabDelete urlPath params = do
+  request
   where
     request =
       gitlabHTTPOne
@@ -120,7 +122,7 @@ gitlabDelete urlPath = do
         "application/x-www-form-urlencoded"
         urlPath
         []
-        []
+        params
 
 -- | Assumes that HTTP error code responses, e.g. 404, 409, won't be
 -- returned as (Left response) value.
@@ -132,14 +134,33 @@ gitlabUnsafe query = do
     Right Nothing -> error "gitlabUnsafe error"
     Right (Just x) -> return x
 
+-- | Lower level query that returns the raw bytestring response from a
+-- GitLab HTTP query. Useful for downloading project archives files.
+gitlabGetByteStringResponse ::
+  -- | the URL to post to
+  Text ->
+  -- | the data to post
+  [GitLabParam] ->
+  GitLab (Response BSL.ByteString)
+gitlabGetByteStringResponse urlPath params =
+  request
+  where
+    request =
+      gitlabHTTP
+        "GET"
+        "application/x-www-form-urlencoded"
+        urlPath
+        params
+        []
+
 ---------------------
 -- internal functions
 
 gitlabHTTP ::
   -- | HTTP method (PUT, POST, DELETE, GET)
-  ByteString ->
+  BS.ByteString ->
   -- | Content type (content-type)
-  ByteString ->
+  BS.ByteString ->
   -- | the URL
   Text ->
   -- | the URL parameters for GET calls
@@ -148,28 +169,30 @@ gitlabHTTP ::
   [GitLabParam] ->
   GitLab (Response BSL.ByteString)
 gitlabHTTP httpMethod contentType urlPath urlParams contentParams = do
-  cfg <- serverCfg <$> ask
-  manager <- httpManager <$> ask
+  cfg <- MR.asks serverCfg
+  manager <- MR.asks httpManager
   let url' = url cfg <> "/api/v4" <> urlPath <> T.decodeUtf8 (renderQuery True urlParams)
+  let authHeader = case token cfg of
+        AuthMethodToken t -> ("PRIVATE-TOKEN", T.encodeUtf8 t)
+        AuthMethodOAuth t -> ("Authorization", "Bearer " <> T.encodeUtf8 t)
   let request' = parseRequest_ (T.unpack url')
       request =
         request'
           { method = httpMethod,
             requestHeaders =
-              [ ("PRIVATE-TOKEN", T.encodeUtf8 (token cfg)),
+              [ authHeader,
                 ("content-type", contentType)
               ],
             requestBody = RequestBodyBS (renderQuery False contentParams)
           }
-  response <- liftIO $ tryGitLab 0 request (retries cfg) manager Nothing
-  return response
+  liftIO $ tryGitLab 0 request (retries cfg) manager Nothing
 
 gitlabHTTPOne ::
   FromJSON a =>
   -- | HTTP method (PUT, POST, DELETE, GET)
-  ByteString ->
+  BS.ByteString ->
   -- | Content type (content-type)
-  ByteString ->
+  BS.ByteString ->
   -- | the URL
   Text ->
   -- | the URL query data for GET calls
@@ -193,9 +216,9 @@ gitlabHTTPOne httpMethod contentType urlPath urlParams contentParams = do
 gitlabHTTPMany ::
   (FromJSON a) =>
   -- | HTTP method (PUT, POST, DELETE, GET)
-  ByteString ->
+  BS.ByteString ->
   -- | Content type (content-type)
-  ByteString ->
+  BS.ByteString ->
   -- | the URL
   Text ->
   -- | the URL query data for GET calls
@@ -221,24 +244,22 @@ gitlabHTTPMany httpMethod contentType urlPath urlParams contentParams = do
           case parseMany (responseBody response) of
             Nothing -> return (Right accum)
             Just moreResults -> do
-              let numPages = totalPages response
-                  accum' = accum <> moreResults
-              if pageNum == numPages
-                then return (Right accum')
-                else go (pageNum + 1) accum'
+              let accum' = accum <> moreResults
+              if hasNextPage response
+                then go (pageNum + 1) accum'
+                else return (Right accum')
         else return (Left response)
 
-totalPages :: Response a -> Int
-totalPages resp =
+hasNextPage :: Response a -> Bool
+hasNextPage resp =
   let hdrs = responseHeaders resp
    in findPages hdrs
   where
-    findPages [] = 1 -- error "cannot find X-Total-Pages in header"
-    findPages (("X-Total-Pages", bs) : _) =
-      case readMaybe (T.unpack (T.decodeUtf8 bs)) of
-        Just s -> s
-        Nothing -> error "cannot find X-Total-Pages in header"
+    findPages [] = False
+    findPages (("X-Next-Page", bs) : _) = isJust $ readNP bs
     findPages (_ : xs) = findPages xs
+    readNP :: BS.ByteString -> Maybe Int
+    readNP bs = readMaybe (T.unpack (T.decodeUtf8 bs))
 
 successStatus :: Status -> Bool
 successStatus (Status n _msg) =
@@ -259,8 +280,8 @@ tryGitLab ::
 tryGitLab i request maxRetries manager lastException
   | i == maxRetries = error (show lastException)
   | otherwise =
-    httpLbs request manager
-      `Exception.catch` \ex -> tryGitLab (i + 1) request maxRetries manager (Just ex)
+      httpLbs request manager
+        `Exception.catch` \ex -> tryGitLab (i + 1) request maxRetries manager (Just ex)
 
 parseOne :: FromJSON a => BSL.ByteString -> Maybe a
 parseOne bs =
